#!/usr/bin/env python3

"""
Do stuff to trees

Usage:
    smot tips [--format=<format>] [<filename>]
    smot plot [--format=<format>] [<filename>]
    smot factor [--format=<format>]
                [--factor-by-field=<factorByField>]
                [--factor-by-capture=<capture>]
                [--factor-by-table=<tablefile>]
                [--default=<defaultFactor>]
                [--impute] [<filename>]
    smot sample-equal [--format=<format>]
                      [--factor-by-field=<factorByField>]
                      [--factor-by-capture=<capture>]
                      [--factor-by-table=<tablefile>]
                      [--default=<defaultFactor>]
                      [--keep=<keep>] [--max-tips=<tips>] [--zero] [<filename>]
    smot sample-proportional [--format=<format>]
                             [--factor-by-field=<factorByField>]
                             [--factor-by-capture=<capture>]
                             [--factor-by-table=<tablefile>]
                             [--default=<defaultFactor>]
                             [--paraphyletic]
                             [--proportion=<proportion>] [--keep=<keep>] [--seed=<seed>]
                             [--min-tips=<tips>] [--zero] [<filename>]
    smot tipsed [--format=<format>] <pattern> <replacement> [<filename>]
    smot midpoint [--format=<format>] [<filename>]
    smot random [--format=<format>] [--seed=seed] [<tipnames>]
    smot clean [<filename>]

Options
    --zero                    Set branches without lengths to 0  
    -f --format STR           Tree format (newick or nexus)
    -k --keep LIST            Factors to keep
    -m --max-tips INT         Maximum number of tips to keep per unkept factor
    --factor-by-field INT     Factor by field index (with '|' delimiters, for now)
    --factor-by-capture REGEX A regular expression with a capture for determining factors from labels
    -p --proportion NUM       The proportion of tips in a clade to keep
    -d --default STR          The name to assign to tips that do not match a factor
    --paraphyletic            Sample across branches
    --impute                  Infer the factor from context, if possible
"""

import signal
import os
from docopt import docopt
from smot.version import __version__


def factorTree(tree, args, default=None):
    import smot.algorithm as alg

    if args["--factor-by-field"]:
        try:
            field = int(args["--factor-by-field"])
        except ValueError:
            die(
                f"""Expected a positive integer for field --factor-by-field, got '{args["--factor-by-field"]}'"""
            )
        tree = alg.factorByField(tree, field, default=default)
    elif args["--factor-by-capture"]:
        tree = alg.factorByCapture(
            tree, pat=args["--factor-by-capture"], default=default
        )
    elif args["--factor-by-table"]:
        tree = alg.factorByTable(
            tree, filename=args["--factor-by-table"], default=default
        )
    tree = alg.setFactorCounts(tree)
    return tree


def cast(args, field, default, lbnd=None, rbnd=None, caster=None, typename=None):
    """
    Cast a field from the command line argument list.
    """
    if args[field]:
        if caster:
            try:
                x = caster(args[field])
                if lbnd and x < lbnd:
                    print(f"Expected {field} to be greater than or equal to {lbnd}")
                if rbnd and x > rbnd:
                    print(f"Expected {field} to be less than or equal to {rbnd}")
            except ValueError:
                print(f"Expected argument {field} to be a {typename}")
                sys.exit(1)
        else:
            x = args[field]
    else:
        x = default
    return x


def main():
    if os.name is "posix":
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)

    args = docopt(__doc__, version=f"smot {__version__}")

    from smot.classes import Node
    from smot.parser import p_newick
    import smot.algorithm as alg
    import sys

    sys.setrecursionlimit(10 ** 8)

    format = cast(args, "--format", "newick")

    if args["random"]:
        from Bio import Phylo
        import random

        if args["<tipnames>"]:
            with open(tipfile, "r") as f:
                names = [name.strip() for name in f.readlines()]
        else:
            names = [name.strip() for name in sys.stdin]
        seed = cast(args, "--seed", None, caster=int, typename="int", lbnd=0)
        random.seed(seed)
        btree = Phylo.BaseTree.Tree.randomized(names)
        Phylo.write(btree, file=sys.stdout, format="newick")
        sys.exit(0)

    if args["<filename>"]:
        f = open(args["<filename>"], "r")
    else:
        f = sys.stdin

    if args["midpoint"]:
        from Bio import Phylo

        tree = list(Phylo.parse(f, format=format))[0]
        tree.root_at_midpoint()
        Phylo.write(tree.clade, file=sys.stdout, format="newick")
        sys.exit(0)
    elif args["plot"]:
        from Bio import Phylo

        btree = list(Phylo.parse(f, format=format))[0]
        Phylo.draw(btree)
        sys.exit(0)

    rawtree = f.readlines()
    rawtree = "".join(rawtree).strip()
    tree = p_newick.parse(rawtree)

    if args["tipsed"]:
        import re

        pat = re.compile(args["<pattern>"])

        def fun_(nodeData):
            if nodeData.label:
                nodeData.label = re.sub(pat, args["<replacement>"], nodeData.label)
            return nodeData

        tree = alg.treemap(tree, fun_)
        print(tree.newick())
    elif args["clean"]:
        tree = alg.clean(tree)
        print(tree.newick())
    elif args["tips"]:
        tree = alg.setNLeafs(tree)

        def _fun(b, x):
            if x.isLeaf:
                b.append(x.label)
            return b

        for tip in alg.treefold(tree, _fun, []):
            print(tip)
    elif args["factor"]:
        defaultFactor = cast(args, "--default", None)
        tree = factorTree(tree, args, default=defaultFactor)
        if args["--impute"]:
            tree = alg.imputeFactors(tree)
        def _fun(b, x):
            if x.isLeaf:
                if x.factor is None:
                    factor = defaultFactor
                else:
                    factor = x.factor
                b.append(f"{x.label}\t{factor}")
            return b

        for row in alg.treefold(tree, _fun, []):
            print(row)

    elif args["sample-equal"] or args["sample-proportional"]:
        defaultFactor = cast(args, "--default", None)
        tree = factorTree(tree, args, default=defaultFactor)
        keep = cast(
            args,
            "--keep",
            [],
            caster=lambda x: x.split(","),
            typename="comma separated list",
        )
        minTips = cast(args, "--min-tips", 3, caster=int, typename="int", lbnd=0)
        maxTips = cast(args, "--max-tips", 5, caster=int, typename="int", lbnd=0)
        seed = cast(args, "--seed", None, caster=int, typename="int", lbnd=0)
        proportion = cast(
            args, "--proportion", 0.5, caster=float, typename="float", lbnd=0, rbnd=1
        )
        if args["sample-equal"]:
            tree = alg.sampleContext(tree, keep=keep, maxTips=maxTips)
        elif args["sample-proportional"] and not args["--paraphyletic"]:
            tree = alg.sampleProportional(
                tree, keep=keep, proportion=proportion, minTips=minTips, seed=seed
            )
        elif args["sample-proportional"] and args["--paraphyletic"]:
            tree = alg.sampleParaphyletic(
                tree, keep=keep, proportion=proportion, minTips=minTips, seed=seed
            )
        print(tree.newick())
    else:
        print(tree.newick())


if __name__ == "__main__":
    main()



import parsec as p
from smot.classes import Node


def p_parens(parser):
    return p.string("(") >> parser << p.string(")")


def p_brackets(parser):
    return p.string("[") >> parser << p.string("]")


def p_tuple(parser):
    return p_parens(p.sepBy1(parser, p.string(",")))


p_number = (
    p.regex("-?\d\.?\d*[eE]-?\d+") ^ p.regex("-?\d+\.\d+") ^ p.regex("-?\d+")
).parsecmap(float)
p_label = (
    (p.string("'") >> p.regex("[^']+") << p.string("'"))
    ^ (p.string('"') >> p.regex('[^"]+') << p.string('"'))
    ^ p.regex("[^,:;()[\]]+")
)
p_length = p.string(":") >> p_number
p_format = p_brackets(p.regex("[^\]]*"))

p_term = (p_label + p.optional(p_format) + p.optional(p_length)).parsecmap(
    lambda x: (x[0][0], x[0][1], x[1])
)

p_leaf = p_term.parsecmap(lambda x: Node(kids=[], label=x[0], form=x[1], length=x[2]))
p_info = (p.optional(p_label) + p.optional(p_format) + p.optional(p_length)).parsecmap(
    lambda x: (x[0][0], x[0][1], x[1])
)


@p.generate
def p_node():
    r = yield (p_tuple(p_leaf ^ p_node) + p.optional(p_info)).parsecmap(
        lambda x: Node(kids=x[0], label=x[1][0], form=x[1][1], length=x[1][2])
    )
    return r


p_newick = p_node << p.string(";")



from smot.classes import Node
from smot.util import log, die
from collections import Counter, defaultdict
import re
import sys
import math
import random


def treemap(node, fun, **kwargs):
    """
    Map a function over the data (label, format, and branch length) of each node in
    the tree. The function is guaranteed to never alter the topology of the tree.

    fun :: NodeData -> NodeData
    """
    node.data = fun(node.data, **kwargs)
    node.kids = [treemap(k, fun, **kwargs) for k in node.kids]
    return node


def treefold(node, fun, init, **kwargs):
    """
    fun :: a -> NodeData -> a
    """
    x = fun(init, node.data, **kwargs)
    for kid in node.kids:
        x = treefold(kid, fun, x, **kwargs)
    return x


def treecut(node, fun, **kwargs):
    """
    fun :: Node -> Node
    """
    node.kids = fun(node, **kwargs)
    node.kids = [treecut(kid, fun, **kwargs) for kid in node.kids]
    return node


def clean(tree, isRoot=True):
    """
    Remove nodes that have only one child. Add the branch lengths.
    """

    def _clean(node, isRoot):
        # remove empty children
        node.kids = [kid for kid in node.kids if kid.data.nleafs > 0]
        # remove all single-child nodes
        while len(node.kids) == 1:
            if node.data.length is not None and node.kids[0].data.length is not None:
                node.kids[0].data.length += node.data.length
            node = node.kids[0]
            # remove empty children
            node.kids = [kid for kid in node.kids if kid.data.nleafs > 0]
        # clean all children
        newkids = []
        for kid in node.kids:
            kid = _clean(kid, False)
            if kid.data.isLeaf or kid.kids:
                newkids.append(kid)
        node.kids = newkids
        # if `tree` is the entire tree and if the tree contains only one leaf, then
        # we need to insert a root node
        if node.data.isLeaf and isRoot:
            node = Node(kids=[node])
        return node

    tree = setNLeafs(tree)
    return _clean(tree, isRoot)


def factorByLabel(node, fun, **kwargs):
    """
    Assign factors to nodes based on the node label string
    """

    def mapfun(ndata, **kwargs):
        ndata.factor = fun(ndata.label, **kwargs)
        return ndata

    return treemap(node, mapfun, **kwargs)


def factorByField(tree, field, default=None, sep="|"):
    """
    Factor by the <field>th 1-based index in the tip label.
    """

    def _fun(name):
        try:
            return name.split(sep)[field - 1]
        except:
            return default

    return factorByLabel(tree, _fun)


def factorByCapture(tree, pat, default=None):
    pat = re.compile(pat)

    def _fun(name):
        if name:
            m = re.search(pat, name)
            if m:
                if m.groups(1):
                    if isinstance(m.groups(1), str):
                        return m.groups(1)
                    else:
                        return m.groups(1)[0]
                else:
                    return m.groups(0)
        return default

    return factorByLabel(tree, _fun)


def factorByTable(tree, filename, default=None):
    with open(filename, "r") as f:
        factorMap = dict()
        for line in f.readlines():
            try:
                (k, v) = line.strip().split("\t")
                factorMap[k] = v
            except ValueError:
                print("Expected two columns in --factor-by-table file", file=sys.stderr)
                sys.exit(1)

        def _fun(name):
            if name:
                for k, v in factorMap.items():
                    if k in name:
                        return v
            return default

        return factorByLabel(tree, _fun)

def isMonophyletic(node):
    """
    Check is a branch is monophyletic relative to the defined factors. Assumes
    that `setFactorCounts` has been called on the tree.
    """
    return len(node.data.factorCount) <= 1


def getFactor(node):
    if node.data.factorCount:
        return list(node.data.factorCount.keys())[0]
    else:
        return None

def imputeFactors(tree):
    def setFactors(node, factor):
        def _fun(b):
            b.factor = factor
            return b
        return treemap(node, _fun)

    if tree.data.factorCount and isMonophyletic(tree):
        tree = setFactors(tree, getFactor(tree))
    else:
        newkids = [imputeFactors(kid) for kid in tree.kids]
        tree.kids = newkids
    return tree

def getLeftmost(node):
    """
    Recurse down a tree, returning the leftmost leaf
    """
    if node.kids:
        return getLeftmost(node.kids[0])
    else:
        return node


def sampleN(node, n):
    if not node.data.nleafs:
        node = setNLeafs(node)
    if n == 0:
        raise "Cannot create empty leaf"
    elif n > node.data.nleafs:
        return node
    if not node.kids and not n == 1:
        raise "Bug in sampleN"
    N = sum([kid.data.nleafs for kid in node.kids])
    selection = distribute(n, len(node.kids), [kid.data.nleafs for kid in node.kids])
    node.kids = [sampleN(kid, m) for kid, m in zip(node.kids, selection) if m > 0]
    if len(node.kids) == 1:
        if node.kids[0].data.length is not None and node.data.length is not None:
            node.kids[0].data.length += node.data.length
        node = node.kids[0]
    return node


def sampleRandom(node, n, rng):
    """
    Sample N random tips from node 
    """
    if not node.data.nleafs:
        node = setNLeafs(node)
    if n == 0:
        raise "Cannot create empty leaf"
    elif n > node.data.nleafs:
        return node
    if not node.kids and not n == 1:
        raise "Bug in sampleN"

    def _collect(b, d):
        if d.isLeaf:
            b.add(d.label)
        return b

    labels = treefold(node, _collect, set())

    # The set datastructure is a hash table. Hashes of strings vary between
    # python sessions. Thus the sorting of elements in a set will also vary
    # between sessions. So before sampling, the set needs to be converted to a
    # list of sorted. Otherwise, the selected labels will be random even given
    # the same random seed.
    chosen = rng.sample(sorted(list(labels)), n)

    def _cull(node):
        chosenOnes = [
            kid
            for kid in node.kids
            if (not kid.data.isLeaf) or (kid.data.label in chosen)
        ]
        return chosenOnes

    sampledTree = treecut(node, _cull)
    sampledTree = clean(sampledTree)
    sampledTree = setNLeafs(sampledTree)
    return sampledTree


def distribute(count, groups, sizes=None):
    """
    Break n into k groups

    For example:

    distribute(5, 3) --> [2, 2, 1]
    distribute(1, 3) --> [1, 0, 0]
    distribute(2, 2) --> [1, 1]

    If the argument 'sizes' is given, then elements are selected from a pool of
    'sizes' things.  Thus there is an upper limit on how many items can be
    selected. For example:

    distributed(10, 3, [1, 4, 4]) --> [1,4,4]
    distributed(10, 3, [1, 5, 5]) --> [1,5,4]
    distributed(5, 3, [1, 1, 1]) --> [1,1,1]
    distributed(5, 3, [1, 1, 1]) --> [1,1,1]
    distributed(5, 3, [0, 10, 1]) --> [0,4,1]
    """
    if not sizes:
        sizes = [count] * groups

    unfilledGroups = sum(s > 0 for s in sizes)

    if count <= unfilledGroups:
        selection = []
        for i in range(groups):
            if count > 0 and sizes[i] > 0:
                selection.append(1)
                count -= 1
            else:
                selection.append(0)
    else:
        selection = []
        for i in range(groups):
            selection.append(min(count // unfilledGroups, sizes[i]))
            sizes[i] -= selection[i]
        remaining = count - sum(selection)
        if remaining > 0 and any(s > 0 for s in sizes):
            remaining_selection = distribute(remaining, groups, sizes)
            selection = [s1 + s2 for s1, s2 in zip(selection, remaining_selection)]
    return selection


def setNLeafs(node):
    if node.data.isLeaf:
        node.data.nleafs = 1
    else:
        node.kids = [setNLeafs(kid) for kid in node.kids]
        node.data.nleafs = sum(kid.data.nleafs for kid in node.kids)
    return node


def setFactorCounts(node):
    if node.data.isLeaf:
        if node.data.factor:
            node.data.factorCount = Counter([node.data.factor])
        else:
            node.data.factorCount = Counter()
    else:
        node.data.factorCount = Counter()
        node.kids = [setFactorCounts(kid) for kid in node.kids]
        node.data.factorCounts = Counter()
        for kid in node.kids:
            node.data.factorCount += kid.data.factorCount

    return node


def sampleContext(tree, keep=[], maxTips=5):
    # recursive sampler
    def _sampleContext(node):
        newkids = []
        for kid in node.kids:
            if (
                len(kid.data.factorCount) == 1
                and list(kid.data.factorCount.values())[0] >= maxTips
            ):
                if list(kid.data.factorCount.keys())[0] in keep:
                    newkids.append(kid)
                else:
                    newkids.append(sampleN(kid, maxTips))
            else:
                newkids.append(_sampleContext(kid))
        node.kids = newkids
        return node

    tree = setNLeafs(tree)
    tree = setFactorCounts(tree)
    return clean(_sampleContext(tree))

def sampleParaphyletic(tree, proportion=0.5, keep=[], minTips=3, seed=None):
    rng = random.Random(seed)

    def getLabels(node):
        def _collect(b, d):
            if d.isLeaf:
                b.add(d.label)
            return b

        return treefold(node, _collect, set())

    def sampleLabels(labels, factor):
        if factor in keep:
            return labels
        else:
            N = min(len(labels), max(minTips, math.ceil(proportion * len(labels))))
            try:
                sample = rng.sample(sorted(list(labels)), N)
            except ValueError:
                log(f"Bad sample size ({N}) for population of size ({len(labels)})")
                sys.exit(1)
            return sample

    def _select(node, selected, paraGroup, paraFactor):
        rebelChild = None
        potentialMembers = []
        canMerge = True
        oldFactor = paraFactor
        for kid in node.kids:
            factor = getFactor(kid)
            if not canMerge:
                potentialMembers.append(kid)
            else:
                if isMonophyletic(kid):
                    if factor == paraFactor or factor is None:
                        potentialMembers.append(kid)
                    elif (
                        factor != paraFactor
                        and factor != oldFactor
                        and paraFactor != oldFactor
                    ):
                        canMerge = False
                        potentialMembers.append(kid)
                    else:
                        selected.update(sampleLabels(paraGroup, paraFactor))
                        paraGroup = getLabels(kid)
                        paraFactor = factor
                else:
                    if rebelChild is None:
                        rebelChild = kid
                    else:
                        canMerge = False
                        selected.update(sampleLabels(paraGroup, paraFactor))
                        paraFactor = None
                        paraGroup = set()
                        selected.update(_select(rebelChild, selected, set(), None))
                        selected.update(_select(kid, selected, set(), None))
        if canMerge and rebelChild is not None:
            for k in potentialMembers:
                paraGroup.update(getLabels(k))
            selected.update(_select(rebelChild, selected, paraGroup, paraFactor))
        else:
            groups = defaultdict(set)
            for k in potentialMembers:
                factor = getFactor(k)
                if isMonophyletic(k):
                    if factor == paraFactor or factor is None:
                        paraGroup.update(getLabels(k))
                    else:
                        groups[factor].update(getLabels(k))
                else:
                    selected.update(_select(k, selected, set(), None))
            selected.update(sampleLabels(paraGroup, paraFactor))
            for (k, v) in groups.items():
                selected.update(sampleLabels(v, factor=k))
        return selected

    tree = setFactorCounts(tree)
    selected = _select(tree, set(), set(), None)

    def _cull(node):
        chosenOnes = [
            kid
            for kid in node.kids
            if (kid.data.isLeaf and (kid.data.label in selected)) or kid.kids
        ]
        return chosenOnes

    sampledTree = treecut(tree, _cull)
    sampledTree = clean(sampledTree)
    return sampledTree


def sampleProportional(tree, proportion=0.5, keep=[], minTips=3, seed=None):
    rng = random.Random(seed)

    def _sample(kid):
        N = max(minTips, math.floor(kid.data.nleafs * proportion))
        return sampleRandom(kid, N, rng=rng)

    # recursive sampler
    def _sampleProportional(node):
        newkids = []
        for kid in node.kids:
            nfactors = len(kid.data.factorCount)
            if nfactors == 0:
                newkids.append(_sample(kid))
            elif nfactors == 1:
                if list(kid.data.factorCount.keys())[0] in keep:
                    newkids.append(kid)
                else:
                    newkids.append(_sample(kid))
            else:
                newkids.append(_sampleProportional(kid))
        node.kids = newkids
        return node

    if not (0 <= proportion <= 1):
        die("Expected parameter 'proportion' to be a number between 0 and 1")
    if not (0 <= minTips and minTips % 1 == 0):
        die("Expected parameter 'minTips' to be a positive integer")
    tree = setNLeafs(tree)
    tree = setFactorCounts(tree)
    return clean(_sampleProportional(tree))



import re
from Bio.Phylo.BaseTree import Clade


class NodeData:
    def __init__(self, label=None, form=None, length=None, factor=None, isLeaf=False):
        self.label = label
        self.form = form
        self.length = length
        self.factor = factor
        self.nleafs = None
        self.factorCount = None
        self.color = None
        self.isLeaf = isLeaf

    def _color(self):
        m = re.match("&!color=(#\d{6})")
        if m:
            self.color = m.groups(1)

    def __eq__(self, other):
        # Equality is based off intrinsic data of the tree, not internal data,
        # such as factor.
        return (
            self.label == other.label
            and self.form == other.form
            and self.length == other.length
        )


class Node:
    index = 0

    def __init__(self, kids=[], **kwargs):
        self.kids = kids
        self.data = NodeData(**kwargs, isLeaf=not bool(kids))

    def newick(self):
        return str(self) + ";"

    def asBiopythonTree(self):
        return Clade(
            branch_length=self.data.length,
            name=self.data.label,
            color=self.data.color,
            clades=[kid.asBiopythonTree() for kid in self.kids],
        )

    def __eq__(self, other):
        return self.kids == other.kids and self.data == other.data

    def __str__(self):

        if self.kids:
            s = "(" + ",".join([str(kid) for kid in self.kids]) + ")"
        else:
            s = ""
        if self.data.label:
            #  s += "x" + str(Node.index)
            #  Node.index += 1
            if set("^,:;()[]'\"").intersection(set(self.data.label)):
                if "'" in self.data.label:
                    if '"' in self.data.label:
                        label = re.sub('"', '\\"', self.data.label)
                        s += f'"{label}"'
                    else:
                        s += f'"{self.data.label}"'
                elif '"' in self.data.label:
                    s += f"'{self.data.label}'"
            else:
                s += self.data.label
        if self.data.form:
            s += "[" + self.data.form + "]"
        if self.data.length is not None:
            s += ":" + "{:0.3g}".format(self.data.length)
        return s

    def __repr__(self):
        return str(self)
